from midiutil.MidiFileGenerator import MidiFileGenerator, MidiTrack
import re
import os
import datetime
import Tkinter
import tkFileDialog
import random
import shutil
import argparse


class FileHandler(object):

    """
    Melody Notation:

    WITHIN PYTHON:
    List of lists of lists followed by an integer.  Each inner-most list in the form
    (Midi Note Number, Duration as 'beat').  The next inner-most lists are considered 'NoteUnits', or groups of notes.
    Midi Note Numbers of 0 represent rests.  Final integer is tempo.
    example: [[[0, 1.0],], [[48, 1.0],], [[51, 0.25], [53, 0.25], [55, 1.0], [51, 0.5],], [[48, 2.0],], 100]

    AS FILE NAME:
    File names for each melody consist of:
        * each note as midi note number *dash* duration (with the '.' replaced by 'p')
        * and each NoteUnit separated by double underscores
        * final integer, separated by triple underscore, is tempo
    The above example becomes:
        0-1p0__48-1p0__51-0p25_52-0p25_55-1p0_51_0p5__48-2p0___100.mid
    """

    def __init__(self):
        self.root_directory = '/Users/cumbojd/personal/music/evolving/output'
        self.directory = None

    def setup_directory(self):
        """Create a folder to store this iteration of mutations."""
        suffix_index = 0
        directory = None
        while directory is None:
            directory = os.path.join(
                self.root_directory,
                "%s_%s" % (str(datetime.datetime.now().date()), suffix_index)
            )
            if not os.path.exists(directory):
                os.makedirs(directory)
                break
            else:
                directory = None
                suffix_index += 1
        # create a 'seed' directory inside the folder to store the file used to generate that day's mutations.
        seed_directory = os.path.join(directory, 'seed')
        if not os.path.exists(seed_directory):
            os.makedirs(seed_directory)
        return directory

    def get_seed_file(self):
        """Get the file name of most recent .mid file from the directory, or open a file dialog if none are present."""
        # get the most recent .mid file...
        most_recent = self.get_most_recent_midi_file(directory=self.root_directory)
        if most_recent:
            return most_recent
        # ...otherwise open a dialog to select the file
        root = Tkinter.Tk()
        root.withdraw()
        file_path = tkFileDialog.askopenfilename()
        file_path_parts = file_path.split('/')
        file_name = file_path_parts[-1]
        # copy the selected file to the output directory so processing can begin
        shutil.copy2(file_path, os.path.join(self.root_directory, file_name))
        return file_name

    @staticmethod
    def get_most_recent_midi_file(directory):
        """Grab the file name of the most recent .mid file in the specified directory.

        Args:
            directory (str): directory to get file from
        """
        os.chdir(directory)
        most_recent = None
        for listed_file in os.listdir("."):
            if listed_file.endswith(".mid"):
                if not most_recent:
                    most_recent = listed_file
                elif os.path.getmtime(listed_file) > os.path.getmtime(most_recent):
                    most_recent = listed_file
        return most_recent

    def copy_seed_file(self, filename):
        """Copy the filename specified to the 'seed' directory of the newly generated directory for the most recent
        mutations."""
        original_path = os.path.join(self.root_directory, filename)
        new_path = os.path.join(self.root_directory, self.directory, 'seed', filename)
        shutil.copy2(original_path, new_path)

    @staticmethod
    def _is_output_folder(directory_name):
        """Determine if the directory is an output folder named in the form 'YYYY-MM-DDX'."""
        regex = re.compile("^[\d]{4}-[\d]{2}-[\d]{2}_[\d]{1,2}$")
        if regex.search(directory_name):
            return True
        return False

    def build_sequence_folder(self):
        """Copy each seed file from each set of mutations into a new folder named 'sequence' and rename each with an
        integer filename describing their place in the series of mutations.  Goal is to have a single folder listing
        each of the mutations in order to be able to view the gradual evolutionary change."""
        os.chdir(self.root_directory)
        output_folders = []
        for item in os.listdir("."):
            if os.path.isdir(item) and self._is_output_folder(item):
                output_folders.append(item)
        sorted_output_folders = sorted(output_folders)
        sequence_directory = os.path.join(self.root_directory, 'sequence')
        # if the folder already exists delete it in case the sequence has become re-ordered somehow
        if os.path.exists(sequence_directory):
            shutil.rmtree(sequence_directory)
        os.makedirs(sequence_directory)
        # Get the seed file from each directory, and save it into the sequence directory, renaming to match the name of
        # the directory it came from ("2014-03-17a.mid")
        for directory in sorted_output_folders:
            seed_directory_path = os.path.join(self.root_directory, directory, 'seed')
            seed_file_name = self.get_most_recent_midi_file(directory=seed_directory_path)
            if seed_file_name:
                original_path = os.path.join(seed_directory_path, seed_file_name)
                new_path = os.path.join(sequence_directory, "%s.mid" % directory)
                shutil.copy2(original_path, new_path)

    @staticmethod
    def filename_to_list(filename):
        """Convert a string into a list representing a melody"""
        filename_regex = re.compile('^([\d]{1,3}\-[\d]+p[\d]{1,2}(_)?(__)?)+___(\d){1,4}\.mid$')
        if not filename_regex.match(filename):
            raise ValueError("File name provided does not fit melody format: %s" % filename)
        melody_list = []
        melody_plus_tempo_string = filename.split('.')[0]
        melody_string, tempo = melody_plus_tempo_string.split('___')
        note_unit_strings = melody_string.split('__')
        for note_unit_string in note_unit_strings:
            note_strings = note_unit_string.split('_')
            note_unit_list = []
            for note_string in note_strings:
                note_values = note_string.split('-')
                note_unit_list.append([int(note_values[0]), float(str(note_values[1]).replace('p', '.'))])
            melody_list.append(note_unit_list)
        melody_list.append(int(tempo))
        return melody_list

    @staticmethod
    def list_to_filename(melody_list):
        """Convert a list representing a melody into a filename-appropriate string"""
        the_string = ""
        tempo = melody_list.pop()
        for note_unit in melody_list:
            if the_string != "":
                the_string += '__'
            the_string += "_".join(["%s-%s" % (note[0], str(note[1]).replace('.', 'p')) for note in note_unit])
        return "%s___%s.mid" % (the_string, tempo)


class MidiMaker(object):
    """Transforms melody lists into MIDI output."""

    def __init__(self, melody, file_handler):
        """
        Args:
            melody (list)
            file_handler (FileHandler)
        """
        copy_of_melody = list(melody)
        self.file_generator = MidiFileGenerator()
        self.tempo = copy_of_melody.pop()
        self.melody = copy_of_melody
        self.track = MidiTrack(channel=1, tempo=self.tempo)
        self.file_handler = file_handler

    def write(self):
        """Write the MIDI file."""
        current_time = 0
        for note_unit in self.melody:
            for note in note_unit:
                if note[0] != 0:  # midi note #s of 0 represent rests; for rests only increment the current_time
                    self.track.add_note(time=current_time, duration=note[1], note=note[0], velocity=100)
                current_time += note[1]
        self.file_generator.tracks.append(self.track)
        self.melody.append(self.tempo)
        if not self.file_handler.directory:
                self.file_handler.directory = self.file_handler.setup_directory()
        self.file_generator.writeToFile(
            os.path.join(
                self.file_handler.directory,
                self.file_handler.list_to_filename(melody_list=self.melody)
            )
        )


class Mutator(object):
    """
    Possible Mutations:
        * Tempo Mutation
        * Pitch Mutations:
            * no change
            * Up half-step
            * Down half-step
            * Up whole-step
            * Down whole-step
            etc...
        * Duration Mutations:
            * no change
            * + .25
            * - .25
            * + .5
            * - .5
            * + 1
            * - 1
        * Splitting Mutations (one note becomes 2 or three; only if duration .5 or greater):
            * If cleanly divisible by 2:
                * split in half
            * If cleanly divisible by 3:
                * split into thirds
        * Deleting Mutations (delete a note from a note_unit, or delete the entire note_unit)
        * Joining Mutations (2 or more notes become one; randomly choose which pitch to use)
        * Adding Mutations (add a note to the note_unit, or add an entirely new note_unit)
    """

    DURATION_CHANGES = {0: 0.5, 1: 1, 2: 0.25}
    SIMPLE_MUTATION_TYPES = [
        ['_mutate_pitch', ],
        ['_mutate_duration', ],
        ['_mutate_pitch', '_mutate_duration']
    ]

    def __init__(
            self,
            seed_melody,
            mutation_percentage=5,
            complex_percentage=18,
            max_tempo_change=10,
            pitch_only=False,
            rhythm_only=False,
    ):
        """
        Args:
            seed_melody (list)
        Kwargs:
            mutation_percentage (int): defines what percent of the time a note_unit should mutate.
            complex_percentage (int): defines what percent of the time a note_unit should do a complex mutation.
            max_tempo_change (int): maximum bpm that the melody's tempo will be altered.
        """
        copy_of_seed_melody = list(seed_melody)  # avoid overwriting original in memory
        self.tempo = copy_of_seed_melody.pop()
        self.max_tempo_change = max_tempo_change
        self.seed_melody = copy_of_seed_melody
        self.mutation_percentage = mutation_percentage
        self.pitch_change_probs = self._build_scaled_probabilities_list(max_change=12)
        self.duration_change_probs = self._build_scaled_probabilities_list(max_change=3)
        self.complex_percentage = complex_percentage
        self.pitch_only = pitch_only  # if True, limit mutations to only pitch changes
        self.rhythm_only = rhythm_only  # if True, limit mutations to only rhythm changes

    @staticmethod
    def _build_scaled_probabilities_list(max_change):
        """Builds a list of integers.  The first is derived by 2 to the power of (max_change - 1).  Each subsequent
        value derived by (previous_value / 2) + previous_value.  This results in a series of values that can be used to
        provide decreasing probabilities for random selections: a random number between 1 and the max value of the list
        is selected.  The index of the lowest value in the list that is greater than or equal to the random value can
        then be used to select the value needed.

        Args:
            max_change (int): the number of values that should be in the resulting list
        Returns:
            List containing Ints that represent scaled probabilities

        Examples:
            max_change=4:
                 1, 2, 3, 4
                 (8, 4, 2, 1)
                 result = 8, 12, 14, 15
            max_change=5:
                 1, 2, 3, 4, 5
                 (16, 8, 4, 2, 1)
                 result = 16, 24, 28, 30, 31
        """

        change_probs = []
        first_value = 2**(max_change - 1)
        change_probs.append(first_value)
        i = max_change
        prob_range = first_value
        prob_value = first_value
        while i > 1:
            prob_range /= 2
            prob_value = prob_range + prob_value
            change_probs.append(prob_value)
            i -= 1
        return change_probs

    def _mutate_duration(self, seed_note):
        """Alter the duration of a given note."""
        note = list(seed_note)  # make a copy of the note so we don't overwrite seed_melody
        while True:  # loop to ensure we ALWAYS alter the duration
            duration_change_rand = random.randint(1, self.duration_change_probs[-1])
            duration_change_index = 0
            for i, dcp_value in enumerate(self.duration_change_probs):
                if duration_change_rand <= dcp_value:
                    duration_change_index = i
                    break
            duration_change_amount = self.DURATION_CHANGES[duration_change_index]
            increase_duration = random.choice([True, False])
            if increase_duration:
                new_duration = note[1] + duration_change_amount
            else:
                new_duration = note[1] - duration_change_amount
            if new_duration > 0:
                note[1] = new_duration
                return note

    def _mutate_pitch(self, seed_note):
        """Alter the pitch of a given note."""
        if seed_note[0] == 0:  # don't attempt to mutate pitch of rests; just return original
            return seed_note
        note = list(seed_note)  # make a copy of the note so we don't overwrite seed_melody.
        pitch_change_rand = random.randint(1, self.pitch_change_probs[-1])
        pitch_change_amount = 0
        for i, pcp_value in enumerate(self.pitch_change_probs):
            if pitch_change_rand <= pcp_value:
                pitch_change_amount = i + 1
                break
        increase_pitch = random.choice([True, False])
        if increase_pitch:
            note[0] += pitch_change_amount
        else:
            note[0] -= pitch_change_amount
        return note

    def _mutate_join(self, note_unit):
        """Join together two notes in a note_unit.

        Args:
            note_unit (list)
        Returns:
            note_unit (list) or None if join can't be performed
        """
        if len(note_unit) == 1:
            return None
        else:
            number_of_notes_to_join = 2
            note_join_probs_list = self._build_scaled_probabilities_list(max_change=len(note_unit) - 1)
            note_join_probs_index = random.randint(1, note_join_probs_list[-1])
            for j, note_change_prob in enumerate(note_join_probs_list):
                if note_join_probs_index <= note_change_prob:
                    number_of_notes_to_join = j + 2
                    break
            notes_to_join = note_unit[0:number_of_notes_to_join]
            remaining_notes = note_unit[number_of_notes_to_join:]
            new_pitch = random.choice(notes_to_join)[0]
            new_duration = 0
            for note in notes_to_join:
                new_duration += note[1]
            remaining_notes.insert(0, [new_pitch, new_duration])
            return remaining_notes

    @staticmethod
    def _mutate_delete(note_unit):
        """Delete a note or convert it to a rest.  May delete entire note_units.

        Args:
            note_unit (list)
        Returns:
            note_unit (list) or None. None implies removal of the entire note_unit.
        """
        convert_to_rest = random.choice([True, False])
        if len(note_unit) == 1:
            if convert_to_rest:
                # return a new note_unit with it's first note's pitch set to 0 (a rest)
                # we create a new one rather than altering the old to avoid overwriting seed melody
                return [[0, note_unit[0][1]]]
            return None  # will result in this note_unit not being included in final list
        else:
            target_note = random.choice(note_unit)
            if convert_to_rest:
                target_note[0] = 0
            else:
                target_note_index = note_unit.index(target_note)
                del note_unit[target_note_index]
            return note_unit

    def _mutate_split(self, note_unit):
        """Split a note in this note_unit into two or three.

        Args:
            note_unit (list)
        Returns:
            note_unit (list), or None if split can't be performed
        """
        target_note = random.choice(note_unit)
        target_note_index = note_unit.index(target_note)
        if target_note[1] < 0.5 or target_note[0] == 0:  # don't split very short notes or rests.
            return None
        elif target_note[1] % 0.75 == 0.0:
            new_duration = target_note[1] / 3
            shortened_note = [target_note[0], new_duration]
            note_unit[target_note_index] = shortened_note
            if self.rhythm_only:
                note_unit.insert(target_note_index + 1, shortened_note)
                note_unit.insert(target_note_index + 2, shortened_note)
            else:
                note_unit.insert(target_note_index + 1, self._mutate_pitch(seed_note=shortened_note))
                note_unit.insert(target_note_index + 2, self._mutate_pitch(seed_note=shortened_note))
            return note_unit
        elif target_note[1] % 0.5 == 0.0:
            shortened_note = [target_note[0], target_note[1] * 0.5]
            note_unit[target_note_index] = shortened_note
            if self.rhythm_only:
                note_unit.insert(target_note_index + 1, seed_note=shortened_note)
            else:
                note_unit.insert(target_note_index + 1, self._mutate_pitch(seed_note=shortened_note))
            return note_unit

    def _mutate_add(self, note_unit):
        """Add a note to a note_unit or an entirely new note_unit.

        Args:
            note_unit (list)
        Returns:
            note_unit (list) or a tuple containing multiple note_units
        """
        target_note = random.choice(note_unit)
        target_note_index = note_unit.index(target_note)
        new_note_unit = random.choice([True, False])
        where_to_add = random.choice(['before', 'after'])
        # do a simple mutation on the target note so it's not too weirdly different from the note_unit
        # that it originated from.
        note_to_add = self._perform_simple_mutation(note=target_note)
        # grab a random value; if below the complex percentage threshold make this new note a rest.
        if random.randint(1, 100) >= self.complex_percentage:
            note_to_add[0] = 0
        if not new_note_unit:
            if where_to_add is 'before':
                insert_at = target_note_index
            else:
                insert_at = target_note_index + 1
            note_unit.insert(insert_at, note_to_add)
            return note_unit
        else:
            if where_to_add is 'before':
                return tuple([[note_to_add, ], note_unit])
            else:
                return tuple([note_unit, [note_to_add, ]])

    def _perform_simple_mutation(self, note):
        """Dynamically call the appropriate simple mutation method(s) for a given note.

        Args:
            note (list)
        Returns:
            note (list)
        """
        # choose which type of mutation (pitch, duration, or both)
        if self.pitch_only:
            mutation_type = ['_mutate_pitch', ]
        elif self.rhythm_only:
            mutation_type = ['_mutate_duration', ]
        else:
            mutation_type = random.choice(self.SIMPLE_MUTATION_TYPES)
        for mutation_method in mutation_type:
            return getattr(self, mutation_method)(note)

    def _mutate_simple(self, note_unit):
        """Perform a 'simple' mutation: 'mutates' a note_unit by altering it's pitch and/or duration.

        Args:
            note_unit (list)
        Returns:
            note_unit (list)
        """
        # if there's more than one note in the note_unit:
        if len(note_unit) > 1:
            # choose which notes in the note_unit to mutate randomly select which to change
            # a probability scale is used to determine how many notes to mutate, with fewer mutations being more common.
            number_of_notes_to_change = 1
            note_change_probs_list = self._build_scaled_probabilities_list(max_change=len(note_unit))
            note_change_index = random.randint(1, note_change_probs_list[-1])
            for j, note_change_prob in enumerate(note_change_probs_list):
                if note_change_index <= note_change_prob:
                    number_of_notes_to_change = j + 1
            notes_to_change = []
            while len(notes_to_change) < number_of_notes_to_change:
                chosen_note = random.choice(note_unit)
                if chosen_note not in notes_to_change:
                    notes_to_change.append(note_unit.index(chosen_note))
        else:
            notes_to_change = [0]
        for note_index in notes_to_change:
            note = note_unit[note_index]
            mutated_note = self._perform_simple_mutation(note=note)
            note_unit[note_index] = mutated_note
        return note_unit

    def _mutate_complex(self, note_unit):
        """Performs a 'complex' mutation: join together, split apart, delete notes from, or add notes to note_units.
        Ensures that there will always be some kind of mutation no matter what.

        Note that some complex mutations are inherently less common as this is currently written.  'split' mutations,
        for example, are not always possible on all note_units, so we'll wind up using some other mutation type
        instead.

        Args:
            note_unit (list)
        Returns:
            note_unit (list) or None if entire note_unit is deleted
        """
        while True:
            operation = random.choice(['join', 'delete', 'split', 'add'])
            if operation == 'join':
                joined_note_unit = self._mutate_join(note_unit=note_unit)
                if joined_note_unit:
                    return joined_note_unit
            elif operation == 'delete':
                note_unit_with_deletion = self._mutate_delete(note_unit=note_unit)
                return note_unit_with_deletion  # may be a list or None
            elif operation == 'split':
                split_note_unit = self._mutate_split(note_unit=note_unit)
                if split_note_unit:
                    return split_note_unit
            else:  # operation == 'add'
                # self._mutate_add() will always return a note_unit, sometimes more than one
                return self._mutate_add(note_unit=note_unit)

        return note_unit

    def mutate_melody(self):
        mutated_melody = []
        # loop through each note_unit
        for note_unit in self.seed_melody:
            try:
                # decide if each one should mutate
                mutate_rand = random.randint(1, 100)
                if mutate_rand <= self.mutation_percentage:
                    mutating_note_unit = list(note_unit)  # copy note_unit so we don't alter seed melody
                    complex_rand = random.randint(1, 100)
                    if complex_rand <= self.complex_percentage and not self.pitch_only:
                        mutation_result = self._mutate_complex(note_unit=mutating_note_unit)
                    else:
                        mutation_result = self._mutate_simple(note_unit=mutating_note_unit)
                    if mutation_result:
                        # in the case of whole note_unit deletions None may be returned
                        # in the case of additions we may have more than one note_unit returned in the mutation_result
                        if isinstance(mutation_result, tuple):
                            for mutated_note_unit in mutation_result:
                                mutated_melody.append(mutated_note_unit)
                        else:
                            mutated_melody.append(mutation_result)
                else:
                    mutated_melody.append(note_unit)
            except TypeError:
                continue
        return mutated_melody

    def mutate_tempo(self):
        mutate_rand = random.randint(1, 100)
        if mutate_rand <= self.mutation_percentage * 2:
            increase_tempo = random.choice([True, False])
            tempo_mutation_amount = 1
            mutate_tempo_probs_list = self._build_scaled_probabilities_list(max_change=self.max_tempo_change)
            mutate_tempo_probs_index = random.randint(1, mutate_tempo_probs_list[-1])
            for j, mutate_tempo_prob in enumerate(mutate_tempo_probs_list):
                if mutate_tempo_probs_index <= mutate_tempo_prob:
                    tempo_mutation_amount += j
                    break
            if increase_tempo:
                return self.tempo + tempo_mutation_amount
            else:
                return self.tempo - tempo_mutation_amount
        else:
            return self.tempo


def main():
    # handle command line arguments
    argparser = argparse.ArgumentParser()
    argparser.add_argument(
        "--pitchonly",
        help="only allow mutations of pitch; all durations/rhythms will stay the same.",
        action="store_true"
    )
    argparser.add_argument(
        "--rhythmonly",
        help="only allow mutations of rhythm; all pitches will stay the same.",
        action="store_true"
    )
    args = argparser.parse_args()
    # begin processing
    file_handler = FileHandler()
    file_handler.directory = file_handler.setup_directory()
    file_handler.build_sequence_folder()
    seed_file = file_handler.get_seed_file()
    file_handler.copy_seed_file(filename=seed_file)
    seed_melody = file_handler.filename_to_list(filename=seed_file)
    print "original melody:"
    print seed_melody
    i = 0
    created_melodies = []
    mutator = Mutator(
        seed_melody=seed_melody,
        mutation_percentage=3,
        pitch_only=args.pitchonly,
        rhythm_only=args.rhythmonly
    )
    while i < 35:
        mutated_melody = mutator.mutate_melody()
        # Don't mutate tempo if pitchonly or rhymonly arguments supplied
        if args.pitchonly or args.rhythmonly:
            new_tempo = mutator.tempo
        else:
            new_tempo = mutator.mutate_tempo()
        mutated_melody.append(new_tempo)
        if mutated_melody != seed_melody and mutated_melody not in created_melodies:
            print "mutated_melody:"
            print mutated_melody
            created_melodies.append(mutated_melody)
            midi_maker = MidiMaker(melody=mutated_melody, file_handler=file_handler)
            midi_maker.write()
            i += 1


if __name__ == "__main__":
    main()